<?php

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\improvements\ImprovementsHelper;
use Drupal\taxonomy\Entity\Term;
use Drupal\taxonomy\TermInterface;
use Drupal\taxonomy\TermStorageInterface;
use Drupal\taxonomy\VocabularyInterface;

/**
 * Implements hook_entity_base_field_info().
 */
function improvements_entity_base_field_info(EntityTypeInterface $entity_type) {
  if ($entity_type->id() == 'taxonomy_term') {
    $fields = [];

    $fields['stored_depth'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Depth'))
      ->setSetting('unsigned', TRUE);

    $fields['has_children'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Has chilren'))
      ->setDefaultValue(FALSE);

    return $fields;
  }
}

/**
 * Implements hook_ENTITY_TYPE_presave(): taxonomy_term.
 */
function improvements_taxonomy_term_presave(TermInterface $term) {
  // Set "stored_depth" field
  $term->set('stored_depth', ImprovementsHelper::getTermDepth($term));

  // Set "has_chilren" field
  if ($term->get('has_children')->isEmpty()) {
    $term->set('has_children', ImprovementsHelper::termHasChildren($term->id()));
  }
}

/**
 * Implements hook_ENTITY_TYPE_insert(): taxonomy_term.
 */
function improvements_taxonomy_term_insert(TermInterface $term) {
  // Set parent term "has_children" field value
  $parent_term = $term->get('parent')->entity; /** @var TermInterface $parent_term */
  if ($parent_term && !$parent_term->get('has_children')->value) {
    $parent_term->set('has_children', TRUE)->save();
  }
}

/**
 * Implements hook_ENTITY_TYPE_update(): taxonomy_term.
 */
function improvements_taxonomy_term_update(TermInterface $term) {
  $original_term = $term->original; /** @var TermInterface $original_term */

  // Manualy save children terms to run improvements_taxonomy_term_presave() and update "stored_depth" field
  $term_depth_is_changed = ($term->get('stored_depth')->value != $original_term->get('stored_depth')->value);
  if ($term_depth_is_changed || $original_term->get('stored_depth')->isEmpty()) {
    if ($term_direct_children = ImprovementsHelper::getTermDirectChildren($term)) {
      foreach ($term_direct_children as $children_term) {
        $children_term->save();
      }
    }
  }

  // Update parent terms "has_children" field
  $term_parent_items = $term->get('parent');
  $original_term_parent_items = $original_term->get('parent');
  $term_parent_is_changed = $term_parent_items->hasAffectingChanges($original_term_parent_items, $term_parent_items->getLangcode());

  if ($term_parent_is_changed) {
    // Update new parent terms
    if ($term_parent_items->entity) {
      foreach ($term_parent_items as $term_parent_item) {
        $parent_term = $term_parent_item->entity; /** @var TermInterface $parent_term */
        if (!$parent_term->get('has_children')->value) {
          $parent_term->set('has_children', TRUE)->save();
        }
      }
    }

    // Update old parent terms
    if ($original_term_parent_items->entity) {
      foreach ($original_term_parent_items as $original_term_parent_item) {
        $original_parent_term = $original_term_parent_item->entity; /** @var TermInterface $original_parent_term */
        $original_parent_term_has_children_new = ImprovementsHelper::termHasChildren($original_parent_term->id());
        if ($original_parent_term->get('has_children')->value != $original_parent_term_has_children_new) {
          $original_parent_term->set('has_children', $original_parent_term_has_children_new)->save();
        }
      }
    }
  }
}

/**
 * Implements hook_ENTITY_TYPE_delete(): taxonomy_term.
 */
function improvements_taxonomy_term_delete(TermInterface $term) {
  // Update parent term "has_children" field value
  $parent_term = $term->get('parent')->entity; /** @var TermInterface $parent_term */
  if ($parent_term && !ImprovementsHelper::termHasChildren($parent_term->id())) {
    $parent_term->set('has_children', FALSE)->save();
  }
}

/**
 * Implements hook_form_BASE_FORM_ID_alter(): taxonomy_term_form.
 */
function improvements_form_taxonomy_term_form_alter(array &$form, FormStateInterface $form_state) {
  // Add autofocus to term name field
  if (isset($form['name'])) {
    $form['name']['widget']['0']['value']['#attributes']['autofocus'] = 'autofocus';
  }

  // Prepopulate parent field based on $_GET['parent'] value
  // @see improvements_entity_operation()
  if (isset($form['relations']['parent']) && ($parent_id = \Drupal::request()->query->get('parent'))) {
    $form['relations']['parent']['#default_value'][0] = $parent_id;
  }
}

/**
 * Implements hook_form_FORM_ID_alter(): taxonomy_overview_terms.
 */
function improvements_form_taxonomy_overview_terms_alter(array &$form, FormStateInterface $form_state) {
  $form['actions']['reset_alphabetical']['#access'] = FALSE;

  // @TODO Replace to VBO tab
  $form['actions']['resave'] = [
    '#type' => 'submit',
    '#value' => t('Resave'),
    '#submit' => ['improvements_form_taxonomy_overview_terms_submit_resave'],
  ];
}

/**
 * "Resave" button submit callback.
 */
function improvements_form_taxonomy_overview_terms_submit_resave(array &$form, FormStateInterface $form_state) {
  $vocabulary = $form_state->get('taxonomy')['vocabulary']; /** @var VocabularyInterface */
  $term_storage = \Drupal::entityTypeManager()->getStorage('taxonomy_term'); /** @var TermStorageInterface $term_storage */
  $terms = $term_storage->loadByProperties(['vid' => $vocabulary->id()]);
  $operations = [];

  foreach ($terms as $term) {
    $operations[] = ['improvements_batch_resave_term', [$term->id()]];
  }

  batch_set([
    'title' => t('Updating'),
    'operations' => $operations,
  ]);
}

/**
 * Batch operation.
 */
function improvements_batch_resave_term(int $term_id) {
  if ($term = Term::load($term_id)) {
    $term->set('stored_depth', NULL);
    $term->set('has_children', NULL);
    $term->save();
  }
}

/**
 * Implements hook_entity_operation().
 */
function improvements_entity_operation(EntityInterface $entity) {
  // Add term operation "Add child term"
  if ($entity->getEntityTypeId() == 'taxonomy_term') {
    $operations = [];

    $url = Url::fromRoute('entity.taxonomy_term.add_form', ['taxonomy_vocabulary' => $entity->bundle()], ['query' => ['parent' => $entity->id()]]);
    $operations['add_child'] = [
      'title' => t('Add child term'),
      'url' => $url,
      'weight' => 10,
    ];

    return $operations;
  }
}
